package org.randombits.support.core.impl;

import com.atlassian.plugin.PluginAccessor;
import com.atlassian.plugin.event.PluginEventManager;
import com.atlassian.plugin.tracker.DefaultPluginModuleTracker;
import com.atlassian.plugin.tracker.PluginModuleTracker;
import org.randombits.support.core.param.*;
import org.randombits.support.core.param.date.DateFormatInterpreter;
import org.randombits.support.core.param.date.DateInterpreter;
import org.randombits.support.core.param.general.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.*;

/**
 * Default implementation of {@link org.randombits.support.core.param.ParameterAssistant}.
 */
public class DefaultParameterAssistant implements ParameterAssistant, DisposableBean {

    private static final Logger LOG = LoggerFactory.getLogger( DefaultParameterAssistant.class );

    private PluginModuleTracker<ParameterInterpreter<?>, ParameterInterpreterModuleDescriptor> interpreterTracker;

    private PluginModuleTracker<ParameterInterceptor, ParameterInterceptorModuleDescriptor> interceptorTracker;

    private final Map<Class<?>, ParameterInterpreter<?>> defaultInterpreters
            = new HashMap<Class<?>, ParameterInterpreter<?>>();

    private final Map<Class<? extends ParameterInterpreter<?>>, ParameterInterpreter<?>> registeredInterpreters
            = new HashMap<Class<? extends ParameterInterpreter<?>>, ParameterInterpreter<?>>();

    private final Map<Class<? extends ParameterInterceptor>, ParameterInterceptor> registeredInterceptors
            = new HashMap<Class<? extends ParameterInterceptor>, ParameterInterceptor>();

    public DefaultParameterAssistant( PluginAccessor pluginAccessor, PluginEventManager pluginEventManager ) {
        registerDefaultInterpreters();

        // Interpreter tracking
        interpreterTracker = new DefaultPluginModuleTracker<ParameterInterpreter<?>, ParameterInterpreterModuleDescriptor>( pluginAccessor, pluginEventManager, ParameterInterpreterModuleDescriptor.class,
                new PluginModuleTracker.Customizer<ParameterInterpreter<?>, ParameterInterpreterModuleDescriptor>() {
                    @Override
                    public ParameterInterpreterModuleDescriptor adding( ParameterInterpreterModuleDescriptor moduleDescriptor ) {
                        registerInterpreter( moduleDescriptor.getModule() );
                        return moduleDescriptor;
                    }

                    @Override
                    public void removed( ParameterInterpreterModuleDescriptor moduleDescriptor ) {
                        unregisterInterpreter( moduleDescriptor.getModule() );
                    }
                } );

        // Interceptor tracking
        interceptorTracker = new DefaultPluginModuleTracker<ParameterInterceptor, ParameterInterceptorModuleDescriptor>( pluginAccessor, pluginEventManager, ParameterInterceptorModuleDescriptor.class, new PluginModuleTracker.Customizer<ParameterInterceptor, ParameterInterceptorModuleDescriptor>() {
            @Override
            public ParameterInterceptorModuleDescriptor adding( ParameterInterceptorModuleDescriptor moduleDescriptor ) {
                registerInterceptor( moduleDescriptor.getModule() );
                return moduleDescriptor;
            }

            @Override
            public void removed( ParameterInterceptorModuleDescriptor moduleDescriptor ) {
                unregisterInterceptor( moduleDescriptor.getModule() );
            }
        } );
    }

    private void registerDefaultInterpreters() {
        registerDefaultInterpreter( new StringInterpreter() );

        registerDefaultInterpreter( new BooleanInterpreter(), boolean.class );

        registerDefaultInterpreter( new ByteInterpreter(), byte.class );
        registerDefaultInterpreter( new ShortInterpreter(), short.class );
        registerDefaultInterpreter( new IntegerInterpreter(), int.class );
        registerDefaultInterpreter( new LongInterpreter(), long.class );
        registerDefaultInterpreter( new DoubleInterpreter(), double.class );
        registerDefaultInterpreter( new NumberFormatInterpreter() );

        registerDefaultInterpreter( new DateInterpreter() );
        registerDefaultInterpreter( new DateFormatInterpreter() );
    }

    private void registerDefaultInterpreter( ParameterInterpreter<?> interpreter, Class<?>... extraTypes ) {
        defaultInterpreters.put( interpreter.getTargetType(), interpreter );

        for ( Class<?> type : extraTypes ) {
            defaultInterpreters.put( type, interpreter );
        }

        registerInterpreter( interpreter );
    }

    private void unregisterInterceptor( ParameterInterceptor interceptor ) {
        registeredInterceptors.put( interceptor.getClass(), interceptor );
    }

    private void registerInterceptor( ParameterInterceptor interceptor ) {
        registeredInterceptors.put( interceptor.getClass(), interceptor );
    }

    private void unregisterInterpreter( ParameterInterpreter<?> interpreter ) {
        registeredInterpreters.remove( interpreter.getClass() );
    }

    private void registerInterpreter( ParameterInterpreter<?> interpreter ) {
        // Not exactly sure why we have to cast here, but it won't compile without it...
        //noinspection unchecked
        registeredInterpreters.put( (Class<? extends ParameterInterpreter<?>>) interpreter.getClass(), interpreter );
    }


    @Override
    public <T extends Parameters> T createParameters( Class<T> paramsType, ParameterSource source ) {
        return paramsType.cast( Proxy.newProxyInstance( paramsType.getClassLoader(), new Class<?>[]{paramsType},
                new ParametersInvocationHandler( paramsType, source ) ) );
    }

    private class ParametersInvocationHandler implements InvocationHandler {

        private final ParameterSource source;

        private final List<Class<? extends ParameterInterceptor>> interceptors;

        private ParametersInvocationHandler( Class<? extends Parameters> paramsType, ParameterSource source ) {
            this.source = source;

            if ( !paramsType.isInterface() )
                throw new IllegalArgumentException( "The parameter type must be an interface." );

            interceptors = new ArrayList<Class<? extends ParameterInterceptor>>();
            findInterceptors( interceptors, paramsType );
        }

        private void findInterceptors( List<Class<? extends ParameterInterceptor>> interceptors, Class<?> paramsType ) {
            Interceptors annotation = paramsType.getAnnotation( Interceptors.class );
            if ( annotation != null ) {
                interceptors.addAll( Arrays.asList( annotation.value() ) );
            }

            for ( Class<?> type : paramsType.getInterfaces() ) {
                if ( Parameters.class.isAssignableFrom( type ) ) {
                    findInterceptors( interceptors, type );
                }
            }
        }

        @Override
        public Object invoke( Object instance, Method method, Object[] params ) throws InterpretationException {
            Class<?> returnType = method.getReturnType();
            Object value = null;
            try {
                value = invoke( instance, method, params, returnType );
            } catch ( InterpretationException e ) {
                if ( supportsException( method, e ) ) {
                    throw e;
                }

                LOG.debug( "InterpretationException while processing an interceptor: " + e.getMessage(), e );
            }
            return value;
        }

        private <T> T invoke( Object instance, Method method, Object[] params, Class<T> outputType ) throws InterpretationException {
            MethodContext<T> context = initContext( instance, method, params, outputType );

            T value = invokeParameter( context );

            return castValue( value, context.getDefaultValue(), outputType );
        }

        @SuppressWarnings("unchecked")
        private <T> T castValue( T value, T defaultValue, Class<T> outputType ) {
            value = value == null ? defaultValue : value;

            if ( value == null && outputType.isPrimitive() ) {
                // Note: We include dodgy casting here due to issues with primitive/Object casting with Class.cast(...).
                if ( boolean.class == outputType ) {
                    return (T) Boolean.FALSE;
                } else {
                    return (T) Integer.valueOf( 0 );
                }
            }
            return value;
        }

        private <T> T invokeParameter( MethodContext<T> context ) throws InterpretationException {
            Method method = context.getMethod();
            String[] paramNames = findParamNames( method );
            for ( String name : paramNames ) {
                String value = source.getValue( name );
                if ( value != null ) {
                    context.setInput( value );
                    if ( interceptors != null ) {
                        for ( Class<? extends ParameterInterceptor> interceptorType : interceptors ) {
                            ParameterInterceptor interceptor = registeredInterceptors.get( interceptorType );
                            if ( interceptor.intercept( context ) == ParameterInterceptor.Result.HALT ) {
                                return context.getReturnValue();
                            }
                        }
                    }

                    ParameterInterpreter<? extends T> interpreter = findInterpreter( method, context.getReturnType() );
                    return interpreter.interpret( context.getInput(), context.getParameterContext() );
                }
            }
            return null;
        }

        private boolean supportsException( Method method, Throwable throwable ) {
            for ( Class<?> type : method.getExceptionTypes() ) {
                if ( type.isInstance( throwable ) )
                    return true;
            }
            return false;
        }

        @SuppressWarnings("unchecked")
        protected <T> MethodContext<T> initContext( Object instance, Method method, Object[] params, Class<T> outputType ) throws InterpretationException {
            MethodContext<T> context = createContext( instance, method, params, outputType );

            Class<?>[] paramTypes = method.getParameterTypes();
            Annotation[][] paramAnnotations = method.getParameterAnnotations();
            boolean defaultSet = false;

            // Check method parameters
            if ( params != null ) {
                for ( int i = 0; i < params.length; i++ ) {
                    Class<?> paramType = paramTypes[i];
                    Object paramValue = params[i];

                    if ( getAnnotation( paramAnnotations[i], Default.class ) != null ) {
                        if ( defaultSet )
                            throw new InterpretationException( "Only one parameter can be marked as the @Default." );

                        if ( outputType.isAssignableFrom( paramType ) ) {
                            defaultSet = true;
                            // Using (T) cast rather than outputType.cast because primitives don't cast well.
                            context.setDefaultValue( (T) paramValue );
                        } else {
                            throw new InterpretationException( "The @Default parameter must be compatible with " + outputType.getName() + ": " + paramTypes[i].getName() );
                        }
                    } else {
                        setContextValue( context.getParameterContext(), paramType, paramValue );
                        Context ctxAnnotation = getAnnotation( paramAnnotations[i], Context.class );
                        if ( ctxAnnotation != null ) {
                            for ( Class<?> type : ctxAnnotation.as() ) {
                                if ( type.isAssignableFrom( paramType ) ) {
                                    setContextValue( context.getParameterContext(), type, paramValue );
                                }
                            }
                        }
                    }
                }
            }

            // Check method annotations
            for ( Annotation annotation : method.getAnnotations() ) {
                if ( annotation.getClass().isAnnotationPresent( Context.class ) ) {
                    setContextValue( context.getParameterContext(), annotation.getClass(), annotation );
                }
            }

            return context;
        }

        private <T> MethodContext<T> createContext( Object instance, Method method, Object[] params, Class<T> returnType ) {
            return new DefaultMethodContext<T>( instance, method, params, returnType, source.createContext() );
        }

        private <T> void setContextValue( ParameterContext context, Class<T> paramType, Object param ) {
            context.set( paramType, paramType.cast( param ) );
        }

        private <T extends Annotation> T getAnnotation( Annotation[] annotations, Class<T> type ) {
            for ( Annotation a : annotations ) {
                if ( type.isInstance( a ) )
                    return type.cast( a );
            }
            return null;
        }

        private <T> ParameterInterpreter<? extends T> findInterpreter( Method method, Class<T> outputType ) throws InterpretationException {
            Interpreter interpreterAnnotation = method.getAnnotation( Interpreter.class );
            ParameterInterpreter<?> interpreter;
            if ( interpreterAnnotation != null ) {
                interpreter = findRegisteredInterpreter( interpreterAnnotation.value() );
            } else {
                interpreter = findDefaultInterpreter( outputType );
            }

            if ( interpreter == null )
                throw new InterpretationException( "Unable to find a valid interpreter for '" + method.getName() );

            // Primitives don't play nice with isAssignableFrom or general casting.
            if ( outputType.isPrimitive() || outputType.isAssignableFrom( interpreter.getTargetType() ) ) {
                //noinspection unchecked
                return (ParameterInterpreter<? extends T>) interpreter;
            } else {
                throw new InterpretationException( "The interpreter's output type of '" + interpreter.getTargetType().getName() + "' is not compatible with the method's return type of '" + outputType.getName() + "'" );
            }
        }

        private ParameterInterpreter<?> findRegisteredInterpreter( Class<? extends ParameterInterpreter<?>> interpreterType ) {
            return registeredInterpreters.get( interpreterType );
        }

        @SuppressWarnings("unchecked")
        private <T> ParameterInterpreter<T> findDefaultInterpreter( Class<T> outputType ) {
            return (ParameterInterpreter<T>) defaultInterpreters.get( outputType );
        }

        private String[] findParamNames( Method method ) {
            // Try the @Parameter annotation first
            Parameter param = method.getAnnotation( Parameter.class );
            if ( param != null )
                return param.value();

            // Otherwise, guess based on the method name.
            String methodName = prepareMethodName( method );
            return new String[]{methodName};
        }

        private String prepareMethodName( Method method ) {
            String methodName = method.getName();
            if ( methodName.startsWith( "get" ) && methodName.length() > 3 ) {
                methodName = methodName.substring( 3, 4 ).toLowerCase() + methodName.substring( 4 );
            } else if ( boolean.class.isAssignableFrom( method.getReturnType() ) && methodName.startsWith( "is" ) && methodName.length() > 2 ) {
                methodName = methodName.substring( 2, 3 ).toLowerCase() + methodName.substring( 3 );
            }
            return methodName;
        }
    }

    @Override
    public void destroy() throws Exception {
        registeredInterpreters.clear();
        interpreterTracker.close();
        interpreterTracker = null;

        registeredInterceptors.clear();
        interceptorTracker.close();
        interceptorTracker = null;
    }
}
